package flowserver

import (
	"encoding/json"
	"errors"
	"flow"
	"flow/registry"
	"fmt"
	"log"
	"reflect"
	"strconv"
	"strings"
	"time"
)

// Node that will contain registry src
type Node struct {
	ID            string            `json:"id"`
	Src           string            `json:"src"`
	Label         string            `json:"label"`
	DefaultInputs map[int]string    `json:"defaultInputs"`
	Prop          map[string]string `json:"prop"`
}

// Link that joins two nodes
type Link struct {
	From string `json:"from"`
	To   string `json:"to"`
	In   int    `json:"in"`
}

// Trigger that join two nodes on state change
type Trigger struct {
	From string   `json:"from"`
	To   string   `json:"to"`
	On   []string `json:"on"`
}

// FlowDocument flow document
type FlowDocument struct {
	Nodes    []Node    `json:"nodes"`
	Links    []Link    `json:"links"`
	Triggers []Trigger `json:"triggers"`
}

func (fd *FlowDocument) fetchTriggerFrom(ID string) []Trigger {
	ret := []Trigger{}
	for _, t := range fd.Triggers {
		if t.From != ID {
			continue
		}
		log.Println("Trigger is:", ID, "adding")
		ret = append(ret, t)
	}
	return ret
}

func (fd *FlowDocument) fetchLinksTo(ID string) []Link {
	ret := []Link{}
	for _, l := range fd.Links {
		if l.To != ID {
			continue
		}
		ret = append(ret, l)
	}
	return ret
}

var ErrLoop = errors.New("Looping through loops is disabled for now")

// FlowBuild build the graph based on an starting node
func FlowBuild(rawData []byte, r *registry.R, startingID string) (*flow.Flow, error) {

	log.Println("Unmarshal document")
	doc := FlowDocument{[]Node{}, []Link{}, []Trigger{}}
	err := json.Unmarshal(rawData, &doc)
	if err != nil {
		return nil, err
	}

	log.Println("Create new flow and add a registry")
	f := flow.New()
	f.SetRegistry(r)

	log.Println("Temporarly map the nodes with ID")
	nodeMap := map[string]Node{}
	for _, n := range doc.Nodes {
		log.Println("Mapping:", n.ID)
		nodeMap[n.ID] = n
	}
	//inputMap := map[string]flow.Operation{}

	nodeTrack := map[string]bool{}

	log.Println("Recursive node matching")
	var fetchInputs func(ID string) error
	fetchInputs = func(ID string) error {
		log.Println("Check for loops")
		if _, ok := nodeTrack[ID]; ok {
			return ErrLoop //fmt.Errorf("[%v] Looping through nodes is disabled:", ID)
		}
		nodeTrack[ID] = true

		// If flow already has ID just return
		if f.HasOp(ID) {
			log.Println("Entry found, continuing")
			return nil
		}

		node, ok := nodeMap[ID]
		if !ok {
			return fmt.Errorf("node not found [%v]", startingID)
		}

		entry, err := r.Entry(node.Src)
		if err != nil {
			return err
		}

		inputs := doc.fetchLinksTo(node.ID)
		param := make([]flow.Data, len(entry.Inputs))
		// Build ops
		for _, l := range inputs {
			from := nodeMap[l.From]

			switch from.Src {
			case "Input":
				inputID, err := strconv.Atoi(from.Prop["input"])
				if err != nil {
					return errors.New("Invalid inputID value, must be a number")
				}
				param[l.In] = f.In(inputID) // By id perhaps
			case "Variable":
				param[l.In] = f.Var(from.ID, from.Prop["init"])
			case "Const":
				// XXX: Automate this in a func
				raw := from.Label
				//raw := from.Prop["value"]
				raw = strings.TrimSpace(raw)
				if raw[0] == '"' { // its a string
					log.Println("Unmashal string")
					var val string
					json.Unmarshal([]byte(raw), &val)
					param[l.In] = val
					log.Println("next Input")
					continue
				}
				newVal := reflect.New(entry.Inputs[l.In])
				err := json.Unmarshal([]byte(raw), newVal.Interface())
				if err != nil {
					// ignore error?
					log.Println("unmarshalling Error", err)
					return err
					//param[l.In] = nil
					//continue
				}
				param[l.In], _ = f.Const(newVal.Elem().Interface())
			default:
				param[l.In] = f.Res(from.ID)
				// Sub process the child
				if err := fetchInputs(from.ID); err != nil {
					return err
				}
			}
		}
		f.DefOp(node.ID, node.Src, param...)

		// Process triggers for this node
		triggers := doc.fetchTriggerFrom(node.ID)
		for _, t := range triggers {
			log.Println("Trigger:", t)
			log.Println("Registering operator", t.To)
			err := fetchInputs(t.To)
			if err != nil {
				log.Println("Error on trigger input", err)
				return err
			}
			// Register the thing here
			f.Hook(flow.Hook{
				Any: func(name string, ID string, triggerTime time.Time, extra ...interface{}) {
					if name != "Error" && name != "Finish" {
						return
					}
					if ID != t.From {
						log.Printf("ID[%v] triggered [%v], I'm t.From: %v", ID, name, t.From)
						return
					}
					exec := false
					for _, o := range t.On {
						if name == o {
							exec = true
							break
						}
					}
					if !exec {
						log.Println("Mismatching trigger, but its a test")
					}

					log.Println("Matching trigger", t.To)
					op := f.Res(t.To)
					go op.Process(name) // Background
				},
			})

			delete(nodeTrack, node.ID)
		}
		return nil
	}
	err = fetchInputs(startingID)
	if err != nil {
		return nil, err
	}

	return f, nil
}
